// SPDX-License-Identifier: MIT
pragma solidity ^0.8.20;

import "@openzeppelin/contracts/token/ERC20/IERC20.sol";
import "@openzeppelin/contracts/token/ERC20/utils/SafeERC20.sol";
import "@openzeppelin/contracts/utils/Pausable.sol";
import "@openzeppelin/contracts/access/Ownable.sol";
import "@openzeppelin/contracts/utils/ReentrancyGuard.sol";
import "@chainlink/contracts/src/v0.8/shared/interfaces/AggregatorV3Interface.sol";
 

contract GoodMarketPrivateSale is ReentrancyGuard, Ownable, Pausable {
    using SafeERC20 for IERC20;

    IERC20 public immutable token;
    address payable public immutable multisigWallet;
    AggregatorV3Interface internal priceFeed;

    struct SaleInfo {
        uint8 currentPhase;
        bool isEnded;
    }
    SaleInfo public saleInfo;

    uint256 public totalRaised;
    uint256[5] private supplies;

    uint256 public constant MIN_CONTRIBUTION = 0.05 ether;
    uint256 public constant MAX_CONTRIBUTION = 10 ether;

    uint256 public constant MAX_PHASES = 5;

    mapping(address => uint256) public contributions;

    event TokensPurchased(address indexed purchaser, uint256 amount);
    event PhaseStarted(uint8 phase);
    event SaleEnded();

    uint256 private constant TOKEN_PRICE_PHASE_1 = 25000000000000000;  // $0.025 * 1e18
    uint256 private constant TOKEN_PRICE_PHASE_2 = 50000000000000000;  // $0.05 * 1e18
    uint256 private constant TOKEN_PRICE_PHASE_3 = 75000000000000000;  // $0.075 * 1e18
    uint256 private constant TOKEN_PRICE_PHASE_4 = 100000000000000000; // $0.10 * 1e18
    uint256 private constant TOKEN_PRICE_PHASE_5 = 125000000000000000; // $0.125 * 1e18

    uint256 private fallbackEthPrice; // Fallback ETH price in USD with 8 decimals

    // Debugging events
    event FallbackPriceUpdated(uint256 oldPrice, uint256 newPrice);
    event DebugGetEthPrice(uint256 price, bool usedFallback, string reason);
    event DebugBuyTokens(uint8 phase, bool isEnded, uint256 contribution, uint256 tokenAmount);
    event DebugCalculateTokenAmount(uint256 weiAmount, uint8 phase, uint256 ethPrice, uint256 usdAmount, uint256 tokenPrice, uint256 tokenAmount);
    event SalePaused(address indexed account);
    event SaleUnpaused(address indexed account);


    // Custom errors for more gas-efficient and informative error handling
    error SaleNotActive(string reason);
    error InvalidContributionAmount(uint256 sent, uint256 min, uint256 max);
    error InvalidTokenAmount(uint256 amount);
    error InsufficientTokenSupply(uint256 requested, uint256 available);
    error TokenTransferFailed(address to, uint256 amount);
    error InvalidPhase(uint8 phase, uint8 maxPhase);
    error PriceFeedError(string reason);



    constructor(
        IERC20 _token,
        address payable _multisigWallet,
        address _priceFeedAddress
    ) Ownable(msg.sender) {
        require(address(_token) != address(0), "Token address cannot be zero");
        require(_multisigWallet != address(0), "Multisig wallet address cannot be zero");
        require(_priceFeedAddress != address(0), "Price feed address cannot be zero");

        token = _token;
        multisigWallet = _multisigWallet;
        priceFeed = AggregatorV3Interface(_priceFeedAddress);

        // Initialize supplies per phase (tokens with 18 decimals)
        supplies[0] = 4_000_000 * 1e18; // Phase 1: 4 million tokens
        supplies[1] = 4_500_000 * 1e18; // Phase 2: 4.5 million tokens
        supplies[2] = 5_000_000 * 1e18; // Phase 3: 5 million tokens
        supplies[3] = 5_500_000 * 1e18; // Phase 4: 5.5 million tokens
        supplies[4] = 6_000_000 * 1e18; // Phase 5: 6 million tokens

        // Initialize the fallback ETH price to $2,400 per ETH (with 8 decimals)
        fallbackEthPrice = 2400 * 1e8; // $2,400 * 1e8
        emit FallbackPriceUpdated(0, fallbackEthPrice);
        // saleInfo.currentPhase = 0;
    }

    function setFallbackEthPrice(uint256 _price) external onlyOwner {
            require(_price > 0, "Price must be positive");
            uint256 oldPrice = fallbackEthPrice;
            fallbackEthPrice = _price;
            emit FallbackPriceUpdated(oldPrice, _price);
    }

    function getFallbackEthPrice() public view returns (uint256) {
        return fallbackEthPrice;
    }
    
    function startNextPhase() external onlyOwner {
        require(!saleInfo.isEnded, "Sale has already ended");
        require(saleInfo.currentPhase < MAX_PHASES, "All phases completed");

        saleInfo.currentPhase++;

        if (saleInfo.currentPhase == MAX_PHASES) {
            saleInfo.isEnded = true;
            emit SaleEnded();
        }

        emit PhaseStarted(saleInfo.currentPhase);
 
    }

       function endSale() external onlyOwner {
        require(!saleInfo.isEnded, "Sale already ended");
        saleInfo.isEnded = true;
        emit SaleEnded();
    }

    // New function to get the current token price for a specific phase
    function getTokenPrice(uint8 phase) public pure returns (uint256) {
        require(phase > 0 && phase <= MAX_PHASES, "Invalid phase");
        if (phase == 1) return TOKEN_PRICE_PHASE_1;
        else if (phase == 2) return TOKEN_PRICE_PHASE_2;
        else if (phase == 3) return TOKEN_PRICE_PHASE_3;
        else if (phase == 4) return TOKEN_PRICE_PHASE_4;
        else return TOKEN_PRICE_PHASE_5;
    }

    // New function to get the total amount raised during the sale
    function getTotalRaised() public view returns (uint256) {
        return totalRaised;
    }

    function buyTokens() public payable nonReentrant whenNotPaused {
            SaleInfo memory info = saleInfo;
            if (info.currentPhase == 0) {
                emit DebugBuyTokens(info.currentPhase, info.isEnded, msg.value, 0);
                revert SaleNotActive("Sale not active!");
            }

            if (info.isEnded) {
                emit DebugBuyTokens(info.currentPhase, info.isEnded, msg.value, 0);
                revert SaleNotActive("Sale has ended!");
            }

            if (msg.value < MIN_CONTRIBUTION || msg.value > MAX_CONTRIBUTION) {
                emit DebugBuyTokens(info.currentPhase, info.isEnded, msg.value, 0);
                revert InvalidContributionAmount(msg.value, MIN_CONTRIBUTION, MAX_CONTRIBUTION);
            }

            uint256 tokenAmount = calculateTokenAmount(msg.value, info.currentPhase);
            if (tokenAmount == 0) {
                emit DebugBuyTokens(info.currentPhase, info.isEnded, msg.value, tokenAmount);
                revert InvalidTokenAmount(tokenAmount);
            }

            if (supplies[info.currentPhase - 1] < tokenAmount) {
                emit DebugBuyTokens(info.currentPhase, info.isEnded, msg.value, tokenAmount);
                revert InsufficientTokenSupply(tokenAmount, supplies[info.currentPhase - 1]);
            }

            supplies[info.currentPhase - 1] -= tokenAmount;
            totalRaised += msg.value; // Update total raised
            contributions[msg.sender] += msg.value;

            bool success = token.transfer(msg.sender, tokenAmount);
            if (!success) {
                emit DebugBuyTokens(info.currentPhase, info.isEnded, msg.value, tokenAmount);
                revert TokenTransferFailed(msg.sender, tokenAmount);
            }



            emit DebugBuyTokens(info.currentPhase, info.isEnded, msg.value, tokenAmount);
            emit TokensPurchased(msg.sender, tokenAmount);
    }

    function calculateTokenAmount(uint256 weiAmount, uint8 phase) public view returns (uint256) {
        require(phase > 0 && phase <= MAX_PHASES, "Invalid phase");

        uint256 ethPrice = getEthPrice(); // Get ETH price in USD with 8 decimals

        // Calculate USD value of the weiAmount
        uint256 usdAmount = (weiAmount * ethPrice) / 1e8; // Adjust for price feed decimals

        uint256 tokenPrice;
        if (phase == 1) tokenPrice = TOKEN_PRICE_PHASE_1;
        else if (phase == 2) tokenPrice = TOKEN_PRICE_PHASE_2;
        else if (phase == 3) tokenPrice = TOKEN_PRICE_PHASE_3;
        else if (phase == 4) tokenPrice = TOKEN_PRICE_PHASE_4;
        else tokenPrice = TOKEN_PRICE_PHASE_5;

        // Calculate token amount: (USD amount * 1e18) / tokenPrice
        uint256 tokenAmount = (usdAmount * 1e18) / tokenPrice;

        // emit DebugCalculateTokenAmount(weiAmount, phase, ethPrice, usdAmount, tokenPrice, tokenAmount);
        return tokenAmount;
    }

    function getEthPrice() internal view returns (uint256) {
        try priceFeed.latestRoundData() returns (
            uint80 /* roundId */,
            int256 price,
            uint256 /*startedAt*/,
            uint256 /*updatedAt*/,
            uint80 /*answeredInRound*/
        ) {
            if (price <= 0) {
                //emit DebugGetEthPrice(fallbackEthPrice, true, "Non-positive price");
                return fallbackEthPrice;
            }
            
         
            //emit DebugGetEthPrice(uint256(price), false, "Price feed successful");
            return uint256(price); // Price with 8 decimals
        } catch {
            //emit DebugGetEthPrice(fallbackEthPrice, true, "Price feed error");
            return fallbackEthPrice;
        }
    }



    function getRemainingSupply(uint8 phase) public view returns (uint256) {
        require(phase > 0 && phase <= MAX_PHASES, "Invalid phase");
        return supplies[phase - 1];
    }

    function withdrawFunds() external onlyOwner {
        uint256 balance = address(this).balance;
        require(balance > 0, "No funds to withdraw");
        (bool sent, ) = multisigWallet.call{value: balance}("");
        require(sent, "Failed to send ETH");
    }

    function withdrawUnsoldTokens() external onlyOwner {
        require(saleInfo.isEnded, "Sale has not ended");
        uint256 unsoldTokens = token.balanceOf(address(this));
        require(unsoldTokens > 0, "No tokens to withdraw");
        token.safeTransfer(owner(), unsoldTokens);
    }

 

    function pauseSale() public onlyOwner {
        _pause();
        emit SalePaused(msg.sender);
    }

    function unpauseSale() public onlyOwner {
        _unpause();
        emit SaleUnpaused(msg.sender);
    }

    receive() external payable {
        buyTokens();
    }
}
